yalmip('clear')
clc
echo on

%*********************************************************
%
% Sum-of-squares decomposition
%
%*********************************************************
%
% This example shows how to solve some simple SOS-problems.
pause
clc

% Define variables
x = sdpvar(1,1);
y = sdpvar(1,1);
z = sdpvar(1,1);
pause

% Define a polynomial to be analyzed
p = 12+y^2-2*x^3*y+2*y*z^2+x^6-2*x^3*z^2+z^4+x^2*y^2;
pause

% and the corresponding constraint
F = set(sos(p));
pause

% Call solvesos to calculate SOS-decomposition.
solvesos(F);
pause

% Extract decomposition
h = sosd(F);
sdisplay(h)
pause

% To obtain more sparse solutions, it can sometimes be 
% beneficial to minimize the trace of the Gramian Q
% in the decomposition p(x) = v(x)'Qv(x)
pause

solvesos(F,[],sdpsettings('sos.traceobj',1));
h = sosd(F);
sdisplay(h)
pause


% Is it really a SOS-decomposition
% If so, p-h'*h=0
pause
sdisplay(p-h'*h)
pause
clc

% What! Failure!? No, numerical issues.
% Remove all terms smaller than 1e-6
clean(p-h'*h,1e-6)
pause

% The largest coefficient in the polynomial
% p-h'*h is displayed in checkset as the 
% primal residual for a SOS constraint
checkset(F)

pause
clc

% As a second example, we solve a somewhat larger problem.
%
% We want to show that the following matrix is co-positive
J = [1 -1  1  1 -1;
    -1  1 -1  1  1;
     1 -1  1 -1  1;
     1  1 -1  1 -1;
    -1  1  1 -1  1];
pause

% It is clearly co-positive if this polynomial 
% is a sum-of-squares
x1 = sdpvar(1,1);x2 = sdpvar(1,1);x3 = sdpvar(1,1);x4 = sdpvar(1,1);x5 = sdpvar(1,1);
z  = [x1^2 x2^2 x3^2 x4^2 x5^2]';
p  = z'*J*z;
pause

solvesos(set(sos(p)))

% Hmm, failure...
%
% Let's multiply the polynomial with the positive definite function sum(x_i^2)
% If this new polynomial is SOS, then the original is
p_new = p*(x1^2+x2^2+x3^2+x4^2+x5^2);

pause
F = set(sos(p_new));
solvesos(F)
checkset(F)
% We found a decomposition, hence p is SOS and J is co-positive
pause
clc

% Let us now solve a parameterized SOS-problem.
%
% A typical application is to find a lower bound
% on the global minima.
%
% This is done by solving
%   
%            max t
% subject to p(x)-t is SOS
pause

% Define p and t 
x = sdpvar(1,1);
y = sdpvar(1,1);
z = sdpvar(1,1);

p = 12+y^2-2*x^3*y+2*y*z^2+x^6-2*x^3*z^2+z^4+x^2*y^2;
t = sdpvar(1,1)
pause

% maximize t subject to SOS-constraint
% solvesos will automatically categorize t as a 
% parametric variable since it is part of the objective
solvesos(set(sos(p-t)),-t);
pause

% Lower bound
double(t)
pause
clc


% Ok, now for some more advanced coding
%
% Given the nonlinear system dxdt=f(x),
% we will try to prove that z'Pz is a lyapunov function
% where z = [x1;x2;x1x2;x1^2;x2^2] and P positive definite
pause

% Define state-variables and system
x = sdpvar(2,1);
f = [-x(1)-x(1)^3+x(2);-x(2)];

pause

% Define z, P, Lyapunov function and derivatives
z = [x(1);x(2);x(1)^2;x(1)*x(2);x(2)^2];
P = sdpvar(length(z),length(z));
V = z'*P*z;
dVdx = jacobian(V,x);
dVdt = dVdx*f;
pause

% Try to prove that dVdt<0, while minimizing
% trace(P), subject to P>0
%
% All parametric variables (i.e. P) are constrained
% hence solvesos will find them automatically
%
% Alternative
% solvesos(F,trace(P),[],P(:));
%
F = set(sos(-dVdt)) + set(P>eye(5));
solvesos(F,trace(P));
pause
clc

% Checking the validity of the SOS-decomposition is easily done
% (checks SOS-decomposition at the current value of P)
checkset(F)
pause


% For safety, we might want to confirm this 
V = z'*double(P)*z;
h = sosd(F(1));
dVdx = jacobian(V,x);
dVdt = dVdx*f;
sdisplay(clean(-dVdt-h'*h,1e-5))
pause

clc
% Finally, make sure to check out the help in the HTML manual
% for more detailed information.
pause

echo off


