function output = bnb(p)
%BNB          General branch-and-bound scheme for conic programs
%
% BNB applies a branch-and-bound scheme to solve mixed integer
% conic programs (LP, QP, SOCP, SDP) and mixed integer geometric programs.
%
% BNB is never called by the user directly, but is called by 
% YALMIP from SOLVESDP, by choosing the solver tag 'bnb' in sdpsettings.
%
% BNB is used if no other mixed integer solver is found, and
% is only useful for very small problems, due to its simple
% and naive implementation.
%
% The behaviour of BNB can be altered using the fields
% in the field 'bnb' in SDPSETTINGS
%
% bnb.branchrule   Deceides on what variable to branch
%                   'max'     : Variable furthest away from being integer
%                   'min'     : Variable closest to be being integer
%                   'first'   : First variable (lowest variable index in YALMIP)
%                   'last'    : Last variable (highest variable index in YALMIP)
%
% bnb.method       Branching strategy
%                   'depth'   : Depth first
%                   'breadth' : Breadth first
%                   'best'    : Expand branch with lowest lower bound
%                   'depthX'  : Depth until integer solution found, then X (e.g 'depthbest')
%
% solver           Solver for the relaxed problems (standard solver tag, see SDPSETTINGS)
%
% inttol           Tolerance for declaring a variable as integer
%
% gaptol           Exit when (upper bound-lower bound)/(1e-3+abs(lower bound)) < gaptol
%
% round            Round variables smaller than bnb.inttol 
%                   
%
% See also SOLVESDP, BINVAR, INTVAR, BINARY, INTEGER, LMI

% Author Johan Lfberg 
% $Id: bnb.m,v 1.7 2005/01/04 11:16:30 johanl Exp $

% ********************************
% INITIALIZE DIAGNOSTICS IN YALMIP
% ********************************
bnbsolvertime = clock; 
showprogress('Branch and bound started',p.options.showprogress);

% ********************************
% We might have a GP : pre-calc
% ********************************
p.nonlinear = find(~(sum(p.monomtable~=0,2)==1 & sum(p.monomtable,2)==1));

% ********************************
% Define infinite bounds
% ********************************
if isempty(p.ub)
    p.ub = repmat(inf,length(p.c),1);
end
if isempty(p.lb)
    p.lb = repmat(-inf,length(p.c),1);
end

% ********************************
% Extract better bounds from model
% ********************************
if ~isempty(p.F_struc)
    [lb,ub,used_rows] = findulb(p.F_struc,p.K);
    if ~isempty(used_rows)        
        used_rows = used_rows(~any(full(p.F_struc(used_rows,1+p.nonlinear)),2));
        if ~isempty(used_rows)
            lower_defined = find(~isinf(lb));
            if ~isempty(lower_defined)
                p.lb(lower_defined) = max(p.lb(lower_defined),lb(lower_defined));
            end
            upper_defined = find(~isinf(ub));
            if ~isempty(upper_defined)
                p.ub(upper_defined) = min(p.ub(upper_defined),ub(upper_defined));
            end   
            p.F_struc(p.K.f+used_rows,:)=[];
            p.K.l = p.K.l - length(used_rows);
        end
    end
end

% *******************************
% PRE-SOLVE (nothing fancy coded)
% *******************************
if isempty(find(isinf([p.ub;p.lb]))) & p.K.l>0 & isempty(p.nonlinear)
    [p.lb,p.ub] = tightenbounds(-p.F_struc(1+p.K.f:p.K.f+p.K.l,2:end),p.F_struc(1+p.K.f:p.K.f+p.K.l,1),p.lb,p.ub,p.integer_variables);  
end

% ********************************
% ADD CONSTRAINTS 0<x<1 FOR BINARY
% ********************************
if ~isempty(p.binary_variables)
    p.ub(p.binary_variables) =  min(p.ub(p.binary_variables),1);
    p.lb(p.binary_variables) =  max(p.lb(p.binary_variables),0);
end

% *******************************
% PERTURBATION OF LINEAR COST
% *******************************
p.corig = p.c;
if nnz(p.Q)~=0
    g = randn('seed');
    randn('state',1253); %For my testing, I keep this the same...   
    % This perturbation has to be better. Crucial for many real LP problems
    p.c = (p.c).*(1+randn(length(p.c),1)*1e-4); 
    randn('seed',g);
end

% *******************************
% We don't need this
% *******************************
p.options.savesolverinput  = 0;
p.options.savesolveroutput = 0;
p.options.saveduals = 0;

% *******************************
% Display logics
% 0 : Silent
% 1 : Display branching
% 2 : Display node solver prints
% *******************************
switch max(min(p.options.verbose,3),0)
case 0
    p.options.bnb.verbose = 0;
case 1
    p.options.bnb.verbose = 1;
    p.options.verbose = 0;
case 2
    p.options.bnb.verbose = 2;
    p.options.verbose = 0;
case 3
    p.options.bnb.verbose = 2;
    p.options.verbose = 1;
otherwise 
    p.options.bnb.verbose = 0;
    p.options.verbose = 0;    
end

% *******************************
% START BRANCHING
% *******************************
[x_min,solved_nodes,lower,upper] = branch_and_bound(p);

% **********************************
% CREATE SOLUTION
% **********************************
output.problem = 0;
if isinf(upper)
    output.problem = 1;
end
if isinf(-lower)
    output.problem = 2;
end
if solved_nodes == p.options.bnb.maxiter
    output.problem = 3;
end
output.solved_nodes = solved_nodes;
output.x            = x_min;
output.infostr      = yalmiperror(output.problem,'BNB');
output.solverinput  = 0;
output.solveroutput =[];
output.D_struc      = [];
output.solvertime   = etime(clock,bnbsolvertime);

function [x_min,solved_nodes,lower,upper] = branch_and_bound(p)

% *******************************
% SET-UP ROOT PROBLEM
% *******************************
p.depth = 0;
p.lower = NaN;
p.x0    = zeros(length(p.c),1);

% *******************************
% Global stuff
% *******************************
upper   = inf;
lower   = NaN;
x_min   = zeros(length(p.c),1);
stack   = [];

% Create function handle to solver
try
    solver = eval(['@' p.solver.subcall]);
catch
    disp(' ');
    disp('The internal branch & bound solver requires MATLAB 6')
    disp('I am too lazy too do the changes to make it compatible')
    disp('with MATLAB 5. If you really need it, contact me...');
    disp(' ');    
    error(lasterr);
end

% *******************************
% INVARIANT PROBLEM DATA
% *******************************
c = p.corig;
Q = p.Q;
f = p.f;
integer_variables = p.integer_variables;
solved_nodes = 0;

gap = inf;
node = 1;

if p.options.bnb.verbose;disp(' Node   Open       Upper        Lower       Gap     Message');end;
while ~isempty(node) & (solved_nodes < p.options.bnb.maxiter) & (isinf(lower) | gap>p.options.bnb.gaptol)
    
    % ********************************************
    % BINARY VARIABLES ARE FIXED ALONG THE PROCESS
    % ********************************************
    binary_variables  = p.binary_variables;
    
    % ********************************************
    % ASSUME THAT WE WON'T FATHOME
    % ********************************************
    keep_digging = 1;
    message = '';
    
    % *************************************
    % SOLVE NODE PROBLEM 
    % *************************************
    if any(p.ub<p.lb)
        x = zeros(length(p.c),1);
        output.x = x;
        output.problem=1;
    else
        relaxed_p = p;
        relaxed_p.integer_variables = [];
        relaxed_p.binary_variables = [];
        output = feval(solver,relaxed_p);       
         
        x = setnonlinear(p,output.x);  
        % **************************************
        % Hmm, don't remember why this fix...
        % **************************************
        if (p.K.l>0) & any(p.F_struc(p.K.f+1:p.K.f+p.K.l,:)*[1;x]<-1e-5)
            output.problem = 1;
        end        
    end
    solved_nodes = solved_nodes+1;
    
    % **************************************
    % THIS WILL BE INTIAL GUESS FOR CHILDREN
    % **************************************
    p.x0 = x;
    
    % *************************************
    % ROOT NODE HEURISTICS (NOTHING CODED)
    % *************************************      
    if output.problem==0 
        [upper1,x_min1] = heuristics(p,x,solver,output);
        if upper1 < upper
            x_min = x_min1;
            upper = upper1;
            [stack,lower] = prune(stack,upper,p.options,solved_nodes);  
        end
    end
    p = adaptivestrategy(p,upper,solved_nodes);

    
    % *************************************
    % ANY INTEGERS? ROUND?
    % *************************************
    non_integer_binary = abs(x(binary_variables)-round(x(binary_variables)))>p.options.bnb.inttol;
    non_integer_integer = abs(x(integer_variables)-round(x(integer_variables)))>p.options.bnb.inttol;
    if p.options.bnb.round
        x(binary_variables(~non_integer_binary)) = round(x(binary_variables(~non_integer_binary)));
        x(integer_variables(~non_integer_integer)) = round(x(integer_variables(~non_integer_integer)));
    end
    non_integer_binary = find(non_integer_binary);
    non_integer_integer = find(non_integer_integer);
    
    % *************************************
    % CHECK FATHOMING POSSIBILITIES
    % *************************************
    feasible = 1;
    switch output.problem
    case 0
        cost = f+c'*x+x'*Q*x;
        if isnan(lower)
            lower = cost;
        end
        
    case {1,12}
        message = 'Infeasible node';
        keep_digging = 0;
        cost = inf;
        feasible = 0;
    case 2
        message = 'Unbounded node';
        cost = -inf;
    otherwise
        message = 'Numerical problems!';
        % This part has to be much more robust
        cost = f+c'*x+x'*Q*x;             
    end
    
    % **************************************
    % YAHOO! INTEGER SOLUTION FOUND
    % **************************************
    if isempty(non_integer_binary) & isempty(non_integer_integer)
        if (cost<upper) & feasible
            if upper==inf
                message = 'First integer solution';
            else    
                message = 'Improved upper bound';
            end            
             x_min = x;                
             upper = cost;                
             [stack,lower] = prune(stack,upper,p.options,solved_nodes);             
        else
            if feasible
                message = 'Integer node';
            end
        end
        p = adaptivestrategy(p,upper,solved_nodes);
        keep_digging = 0;
    end
    if cost>upper*(1-1e-6)
        if keep_digging==1   
            message = 'Poor lower bound';
        else           
            if feasible
                message = 'Poor node';
            end
        end
        keep_digging = 0;
    end   
    % **********************************
    % CONTINUE SPLITTING?
    % **********************************
    if keep_digging & (cost<upper)

        % **********************************
        % BRANCH VARIABLE
        % **********************************
        [index,whatsplit] = branchvariable(x,integer_variables,binary_variables,p.options,x_min);
     
        % **********************************
        % CREATE NEW PROBLEMS
        % **********************************
        switch whatsplit
        case 'binary'
            [p0,p1] = binarysplit(p,x,index,cost); 
        case 'integer'
            [p0,p1] = integersplit(p,x,index,cost,x_min);         
        otherwise
        end
        
        % **********************************
        % Only save varying data in the tree
        % **********************************
        node1.lb = p1.lb;
        node1.ub = p1.ub;
        node1.depth = p1.depth;
        node1.lower = p1.lower;
        node1.x0 = p1.x0;
        node1.binary_variables = p1.binary_variables;
        node0.lb = p0.lb;
        node0.ub = p0.ub;
        node0.depth = p0.depth;
        node0.lower = p0.lower;
        node0.x0 = p0.x0;        
        node0.binary_variables = p0.binary_variables;        
        stack = push(stack,node1);
        stack = push(stack,node0);         
    end
    
    if ~isempty(stack)
        lower = min([stack.lower]);
    end
 
    % **********************************
    % Get a new node to solve
    % **********************************
    [node,stack] = pull(stack,p.options.bnb.method,x_min,upper); 
    if ~isempty(node)
        p.lb = node.lb;
        p.ub = node.ub;
        p.depth = node.depth;
        p.lower = node.lower;
        p.x0 = node.x0;
        p.binary_variables = node.binary_variables;
    end    
    gap = abs((upper-lower)/(1e-3+abs(lower)));
    if isnan(gap)
        gap = inf;
    end

    if p.options.bnb.verbose;showprogress([num2str2(solved_nodes,3) '     ' num2str2(length(stack)+length(node),2)  '      ' num2str2(upper,6,5) '      ' num2str2(lower,7,5) '     ' num2str2(100*(upper-lower)/(1+abs(lower)),5,3) '%    ' message],p.options.bnb.verbose);end
end
if p.options.bnb.verbose;showprogress([num2str2(solved_nodes,3)  ' Finishing.  Cost: ' num2str(upper) ],p.options.bnb.verbose);end



function stack = push(stackin,p)
if ~isempty(stackin)
    stack = [p;stackin];
else
    stack(1)=p;
end

function [p,stack] = pull(stack,method,x_min,upper);

if ~isempty(stack)
    switch method
    case {'depth','depthfirst','depthbreadth','depthproject','depthbest'}
        [i,j]=max([stack.depth]);
        p=stack(j);  
        stack = stack([1:1:j-1 j+1:1:end]);
        
    case 'breadth'
        [i,j]=min([stack.depth]);
        p=stack(j);  
        stack = stack([1:1:j-1 j+1:1:end]); 
        
    case 'best'
        [i,j]=min([stack.lower]);
        p=stack(j);  
        stack = stack([1:1:j-1 j+1:1:end]);  
        
    otherwise        
    end
else
    p = [];
end



% **********************************
% BRANCH VARIABLE
% **********************************
function [index,whatsplit] = branchvariable(x,integer_variables,binary_variables,options,x_min,Weight)
all_variables = [integer_variables(:);binary_variables(:)];
switch options.bnb.branchrule
case 'first'
    index = min(find(abs(x(all_variables)-round(x(all_variables)))>options.bnb.inttol));
case 'last'
    index = max(find(abs(x(all_variables)-round(x(all_variables)))>options.bnb.inttol));        
case 'min'
    nint = find(abs(x(all_variables)-round(x(all_variables)))>options.bnb.inttol);
    [val,index] = min(abs(x(nint)));    
    index = nint(index);
case 'max'
    [val,index] = max(abs(x(all_variables)-round(x(all_variables))));        
otherwise
    error
end
if index<=length(integer_variables)
    whatsplit = 'integer';
else
    index = index-length(integer_variables);
    whatsplit = 'binary';
end

% **********************************
% SPLIT PROBLEM
% **********************************
function [p0,p1] = binarysplit(p,x,index,lower,options)
p0 = p;
p1 = p;
variable = p.binary_variables(index);

p0.ub(variable)=0;
p0.lb(variable)=0;

p0.lower = lower;
p0.depth = p.depth+1;
p0.binary_variables = setdiff(p0.binary_variables,variable);

p1.ub(variable)=1;
p1.lb(variable)=1;

p1.binary_variables = setdiff(p1.binary_variables,variable);
p1.lower = lower;
p1.depth = p.depth+1;

% % *****************************
% % PROCESS MOST PROMISING FIRST
% % (p0 in top of stack)
% % *****************************
if x(variable)>0.5
    pt=p1;
    p1=p0;
    p0=pt;
end

function [p0,p1] = integersplit(p,x,index,lower,options,x_min)

variable = p.integer_variables(index);
current = x(p.integer_variables(index));
lb = floor(current)+1;
ub = floor(current);

% xi<ub
p0 = p;
p0.lower = lower;
p0.depth = p.depth+1;
p0.x0(variable) = ub;
p0.ub(variable)=min(p0.ub(variable),ub);

% xi>lb
p1 = p;
p1.lower = lower;
p1.depth = p.depth+1;
p1.x0(variable) = lb;
p1.lb(variable)=max(p1.lb(variable),lb);

% *****************************
% PROCESS MOST PROMISING FIRST
% *****************************
if lb-current<0.5
    pt=p1;
    p1=p0;
    p0=pt;
end

function [upper,x_min] = heuristics(p,x0,solver,output)
% Did the root find a solution?
if (output.problem ~= 1) & (output.problem ~= 12)
    x = output.x;
    intvars = [p.integer_variables(:);p.binary_variables(:)];
    x_min = 0*x;
    upper = inf;
    xtemp = x;xtemp(intvars) = round(xtemp(intvars));
    xtemp = setnonlinear(p,xtemp);  
    res = resids(p,xtemp);
    if res>-1e-7
        x = xtemp;
        x_min = x;
        upper = p.f+x'*p.Q*x + p.corig'*x;
        return
    end
    xtemp = x;xtemp(intvars) = fix(xtemp(intvars));
    xtemp = setnonlinear(p,xtemp);  
    res = resids(p,xtemp);
    if res>-1e-7
        x = xtemp;
        x_min = x;
        upper = p.f+x'*p.Q*x + p.corig'*x;
        return
    end
    xtemp = x;xtemp(intvars) = ceil(xtemp(intvars));
    xtemp = setnonlinear(p,xtemp);  
    res = resids(p,xtemp);
    if res>-1e-7
        x = xtemp;
        x_min = x;
        upper = p.f+x'*p.Q*x + p.corig'*x;
        return
    end
    xtemp = x;xtemp(intvars) = floor(xtemp(intvars));
    xtemp = setnonlinear(p,xtemp);  
    res = resids(p,xtemp);
    if res>-1e-7
        x = xtemp;
        x_min = x;
        upper = p.f+x'*p.Q*x + p.corig'*x;
        return
    end
else
    upper = inf;
    x_min = output.x*0;
end


function s = num2str2(x,d,c);
if nargin==3
    s = num2str(x,c);
else
    s = num2str(x);
end
s = [repmat(' ',1,d-length(s)) s];


function res = resids(p,x)
res= [];
if p.K.f>0
    res = -abs(p.F_struc(1:p.K.f,:)*[1;x]); 
end
if p.K.l>0
    res = [res;p.F_struc(p.K.f+1:p.K.f+p.K.l,:)*[1;x]];
end
if (length(p.K.s)>1) | p.K.s>0
    top = 1+p.K.f+p.K.l;
    for i = 1:length(p.K.s)
        n = p.K.s(i);
        X = p.F_struc(top:top+n^2-1,:)*[1;x];top = top+n^2;
        X = reshape(X,n,n);
        res = [res;min(eig(X))];
    end
end


function [stack,lower] = prune(stack,upper,options,solved_nodes)
% *********************************
% PRUNE STACK W.R.T NEW UPPER BOUND
% *********************************
if ~isempty(stack)
    toolarge = find([stack.lower]>upper*(1-1e-4));
    if ~isempty(toolarge)
      %  if options.bnb.verbose;showprogress([num2str2(solved_nodes,3) ' Pruned ' num2str(length(toolarge))  '  nodes'],options.bnb.verbose-1);end
        stack(toolarge)=[];
    end
end
if ~isempty(stack)
    lower = min([stack.lower]);
else
    lower = upper;
end

function x = setnonlinear(p,x)
for i = 1:length(p.nonlinear)
    [vals,pos] = find(p.monomtable(p.nonlinear(i),:));
    z = 1;
    for j = 1:length(pos)
        z = z*x(pos(j))^p.monomtable(p.nonlinear(i),pos(j));
    end
    x(p.nonlinear(i)) = z;
end



function p = adaptivestrategy(p,upper,solved_nodes)
% **********************************'
% SWITCH NODE SELECTION STRATEGY?
% **********************************'        
if strcmp(p.options.bnb.method,'depthproject') & (upper<inf)
    p.options.bnb.method = 'project';
   % if p.options.bnb.verbose;showprogress([num2str2(solved_nodes,3) ' Switching to projected cost search.'],p.options.bnb.verbose);end;
end
if strcmp(p.options.bnb.method,'depthbest') & (upper<inf)
    p.options.bnb.method = 'best';
   % if p.options.bnb.verbose;showprogress([num2str2(solved_nodes,3) ' Switching to best-first search.'],p.options.bnb.verbose);end;
end
if strcmp(p.options.bnb.method,'depthbreadth') & (upper<inf)
    p.options.bnb.method = 'breadth';
   % if p.options.bnb.verbose;showprogress([num2str2(solved_nodes,3) ' Switching to breadth-first search.'],p.options.bnb.verbose);end;
end
if strcmp(p.options.bnb.method,'depthest') & (upper<inf)
    p.options.bnb.method = 'est';
   % if p.options.bnb.verbose;showprogress([num2str2(solved_nodes,3) ' Switching to estimated cost search.'],p.options.bnb.verbose);end;
end   


