function [w,problem] = ipqp(Q,c,C,d,A,b,mu_tol,feas_tol)
% Solves Quadratic programming w'Qw+2c'w, Cw<=d, Aw=b
% using primal-dual interior point method (Nocedal-Wright)
%
% Author : Johan Lfberg, Linkpings universitet

% $Id: ipqp.m,v 1.4 2004/07/06 14:09:36 johanl Exp $

if nargin<8
    feas_tol = 1e-8;
end
if nargin<7
    mu_tol = 1e-8;
end

if isempty(Q)
    Q = spalloc(length(c),length(c),0);
else
    Q = Q/2;
end

density = nnz(C)/numel(C);
if (density < 0.1) | (numel(C)>1e5 & density < 0.5)
    C = sparse(C);
else
    C = full(C);
end

% Just for safety
c = c(:);

% Re-order problem
if ~isempty(C)
    h = symmmd(Q+C'*C);
    hinv = h;
    for i = 1:length(h)
        hinv(i) = find(h==i);
    end
else
    h = 1:length(c);
    hinv = h;
end

Q = Q(h,h);
c = c(h);

if ~isempty(C)
    C = C(:,h);
end

% Projects on equality constraints
if ~isempty(A)
    % A feasible
    w_feas = A\b;
    if (norm(A*w_feas-b)>1e-12)
        error('Linear constraints inconsistent.');
    end
    % Project on feasible directions
    H = null(A);
    c = H'*c;
    Q = H'*Q*H;
    d = d-C*w_feas;
    C = C*H;
else
    H = eye(size(Q,1));
    w_feas = zeros(size(Q,1),1);
end

if isempty(C)
    w = -Q\c;
    if ~isempty(A)
    w = w_feas+H*w;
    end
    w = w(hinv);
    problem = 0;
    return
end

% Number of primal variables
n_w = length(Q);

% Number of dual variables and constraints
n_l = length(d);
n_t = length(d);

% Vector with [1,1,1,1,,1...]
e = ones(n_l,1);

% Initial guess
w = zeros(n_w,1);
t = ones(n_t,1);
lambda = ones(n_l,1);

% Complementary slackness gap
mu = lambda'*t/n_t;
iter = 1;
go_on = 1;
data_norm = max([norm(Q,'inf') norm(C,'inf') norm(d,'inf')]);
initial_feasibility = [];
C_is_sparse =  issparse(C);

r_q = -(Q*w+C'*lambda+c);
r_c = C*w+t-d;
r_t = -t.*lambda;
feasibility = norm([r_q;r_c;r_t],'inf');
initial_feasibility = feasibility;


while go_on

   fprintf(['%3d  %8.4e %8.4e  %8.4e \n'],iter,w'*Q*w+c'*w,mu,feasibility);

   lovert = lambda./t;
   
   % Reduced linear system  
   CT = C';
   for i = 1:n_t
       CT(:,i) = lovert(i)*CT(:,i);
   end
   LeftHand  = Q+CT*C;
   RightHand = r_q-C'*((1./t).*(lambda.*r_c+r_t));

    % Solve
    [R,p] = chol(LeftHand);f = 1e-12;
    while p %& f<100
        [R,p] = chol(LeftHand+eye(length(LeftHand)));f = f*1000;
    end    
    if p
        dw_aff = LeftHand\RightHand;
    else
        dw_aff = R\(R'\RightHand);
    end
    
    % Back substitution    
    dl_aff = r_t./t + lovert.*(r_c+C*dw_aff);
    dt_aff = (r_t-t.*dl_aff)./lambda;

    % Prepare for a corrector step
    % Find suitable steplength
    alpha = 0.99999*pdsteplength([lambda;t],[dl_aff;dt_aff],1);

    % Mehrotra heuristic weight
    mu_aff = (lambda+alpha*dl_aff)'*(t+alpha*dt_aff)/n_t;
    sigma = (mu_aff/mu)^3;

    % ****************
    % Corrector step
    % ****************
    r_t = -dl_aff.*dt_aff+sigma*mu*e;
    RightHand = -C'*(r_t./t);
    % Solve
    if p
        dw_cen = LeftHand\RightHand;
    else
        dw_cen = R\(R'\RightHand);
    end
    % Back substitute
    dl_cen = lovert.*(r_t./lambda+C*dw_cen);
    dt_cen = (r_t-t.*dl_cen)./lambda;

    % Combine steps
    d_w = dw_cen+dw_aff;
    d_l = dl_cen+dl_aff;
    d_t = dt_cen+dt_aff;

    % Find suitable steplength
    alpha = 0.99999*pdsteplength([lambda;t],[d_l;d_t],1);    

    % New iterates
    w = w+alpha*d_w;
    t = t+alpha*d_t;
    lambda = lambda+alpha*d_l;
  
    r_q = -(Q*w+C'*lambda+c);
    r_c = (C*w+t-d);
    r_t = -t.*lambda;
    feasibility = norm([r_q;r_c;r_t],'inf');
  
    % New complemetary gap
    mu = lambda'*t/n_t;
    obj_k(iter) = w'*Q*w+c'*w;
    gap_k(iter) = abs(obj_k(iter)+lambda'*d);
    phi_k(iter) = (feasibility+gap_k(iter))/data_norm;

    infeasible = (phi_k(iter) > 1e-8) & phi_k(iter)>1e4*min(phi_k);
    unbounded = all(C*w-d <= 0) & obj_k(iter)<-1e10 ;
    unknown =  (feasibility > 1e-8*data_norm & (feasibility/mu > 1e8*initial_feasibility));
    if iter > 30
         unknown = unknown | (min(phi_k) > 0.5*min(phi_k(2:end-29)));
    end
    success = ~(unknown | unbounded | infeasible) & ((mu < mu_tol) & (feasibility < feas_tol*data_norm));

    go_on = (iter < 50) & ~(unbounded | success | infeasible); %~infeasible & ~unbounded & (feasibility>1e-12 | mu>1e-6);
    iter = iter+1;

end


% Go back to original variables
if ~isempty(A)
    w = w_feas+H*w;
end
w = w(hinv);

% Status code
problem = 0;
if infeasible
    problem = 1;
elseif unbounded
    problem = 2;
end



